import re

from .cast import ensure_data_types
from .cond_extractor import prepare_cond_template
from .var_extractor import prepare_var_template
from . import for_loop


ESCAPE_CHARS = ['(', ')', '?', '.']


def escape_char(tpl):
    for char in ESCAPE_CHARS:
        tpl = tpl.replace(char, f'\\{char}')
    return tpl

def add_border(tpl):
    return f'^{tpl}$'

def prepare_regex(regex, loops):
    regex = for_loop.extract_top(regex, loops)
    regex = prepare_var_template(regex)
    regex = prepare_cond_template(regex)
    return regex

def extract(template, input_):
    regex = escape_char(template)
    loops = for_loop.find_top(regex)

    regex = prepare_regex(regex, loops)
    regex = add_border(regex)

    data = ensure_data_types(re.search(regex, input_, re.DOTALL).groupdict())
    data = extract_data_from_loops(loops, data)
    return data

def extract_data_from_loop(loop, data):
    regex = loop.content
    loops = for_loop.find_top(regex)
    regex = prepare_regex(regex, loops)

    local_data_per_iteration = [
        item.groupdict()
        for item in re.finditer(regex, data[loop.id])
    ]

    loop_variables = {
        loop.name: {
            local_data[loop.vars[0]]: local_data[loop.vars[1]]
            for local_data in local_data_per_iteration
        } if loop.is_dict else [
            local_data[loop.vars[0]]
            for local_data in local_data_per_iteration
        ]
    }

    variables_unrelated_to_loop = {
        key: value
        for key, value in local_data_per_iteration[0].items()
        if key not in loop.vars
    }

    data_with_nested_loops_data = extract_data_from_loops(loops, data)

    return {
        **data_with_nested_loops_data,
        **ensure_data_types(loop_variables),
        **ensure_data_types(variables_unrelated_to_loop)
    }

def extract_data_from_loops(loops, data):
    for loop in loops:
        data = extract_data_from_loop(loop, data)
    return data



import re

#{% for <key_name>, <value_name> in <dict_name>.items() %}<content>{% endfor %}

COND_TYPE_PREFIX = '__cond_'

BEFORE_IF_REGEX = r'{%\s*if\s+(?P<if_name>\w+)\s*%}'
AFTER_IF_REGEX = r'{%\s*endif\s*%}'

def prepare_cond_template(template):
    template = re.sub(BEFORE_IF_REGEX, rf'(?P<{COND_TYPE_PREFIX}\1>', template, 0, re.DOTALL)
    template = re.sub(AFTER_IF_REGEX, ')?', template, 0, re.DOTALL)
    return template

def is_cond_key(key):
    return key.startswith(COND_TYPE_PREFIX)

def cast_cond_key(key):
    return key[len(COND_TYPE_PREFIX)::]

def correct_key(key):
    return cast_cond_key(key) if is_cond_key(key) else key

def correct_value(key, value):
    return bool(value) if is_cond_key(key) else value

def cast_conds(data):
    return {
        correct_key(key): correct_value(key, val)
        for key, val in data.items()
    }



import sonotoria

from .test_extract_variable import load_test_file


def test_extract_list():
    # Given
    template = load_test_file('test_extract_list_template.txt')
    input_ = load_test_file('test_extract_list_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['parameters'] == [
        'lol',
        'bar',
        3.14,
        True
    ]


def test_extract_list_template_multiline():
    # Given
    template = load_test_file('test_extract_list_multiline_template.txt')
    input_ = load_test_file('test_extract_list_multiline_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['parameters'] == [
        'lol',
        'bar',
        3.14,
        True
    ]



# -*- coding: utf-8 -*-
from typing import Any
from collections.abc import Callable
import yaml

from .jinja import template_string

def load(
    path: str,
    filters: dict[str, Callable[[Any], Any]]  = None,
    tests:   dict[str, Callable[[Any], bool]] = None,
    types:   dict[str, object]                = None,
    context: dict[str, Any]                   = None
) -> Any:
    """This function reads the file at the given path, templates it using jinja2 then returns the yaml data.

    Args:
        path: The path of the yaml file.
        filters: A dict with filters to add to the template engine. Names as keys, functions as values.
        tests: A dict of tests to add to the template engine. Names as keys, functions as values.
        types: A dict of objects to add to the template engine. Yaml tags as keys, objects as values.
        context: A dict with additional value for the context used when loading the yaml.

    Returns:
        The data contained in the templated yaml. Most probably a dict.

    Examples:
       >>> from sonotoria import yaml
       >>> data = yaml.load('my_file.yml')
    """
    with open(path, encoding='utf-8') as file:
        lines = file.read()
    return loads(lines, filters, tests, types, context)

def loads(
    yaml_content: str,
    filters: dict[str, Callable[[Any], Any]]  = None,
    tests:   dict[str, Callable[[Any], bool]] = None,
    types:   dict[str, object]                = None,
    context: dict[str, Any]                   = None
) -> Any:
    """This function reads the file at the given path, templates it using jinja2 then returns the yaml data.

    Args:
        yaml_content: The yaml as a string.
        filters: A dict with filters to add to the template engine. Names as keys, functions as values.
        tests: A dict of tests to add to the template engine. Names as keys, functions as values.
        types: A dict of objects to add to the template engine. Yaml tags as keys, objects as values.
        context: A dict with additional value for the context used when loading the yaml.

    Returns:
        The data contained in the templated yaml. Most probably a dict.

    Examples:
       >>> from sonotoria import yaml
       >>> yaml.loads('---\nparam: value\nparam2: {{ param }}')
       {'param': 'value', 'param2': 'value'}
    """
    for tag, type_ in (types or {}).items():
        _add_data_type(tag, type_)

    current_context = {}

    content = []
    for line in yaml_content.split('\n'):
        if '{{' not in line:
            content.append(line)
        else:
            current_context = yaml.safe_load('\n'.join(content))
            content.append(
                template_string(
                    line,
                    merge_context(current_context, context),
                    filters = filters or {},
                    tests = tests or {}
                )
            )

    return yaml.safe_load('\n'.join(content))

def merge_context(context1, context2):
    if isinstance(context1, dict) and isinstance(context2, dict):
        return context1 | context2
    if isinstance(context2, dict) and context1 is None:
        return context2
    return context1

class MyDumper(yaml.Dumper): #pylint: disable=too-many-ancestors
    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)

def dumps(data):
    return yaml.dump(data, Dumper=MyDumper, default_flow_style=False)

def dump(file, data):
    with open(file, encoding='utf-8'):
        file.write(dumps(data))

def ordered(class_):
    yaml.add_representer(class_, _represent_ordered)
    return class_

def constructed(class_):
    yaml.SafeLoader.add_constructor(class_.yaml_tag, class_.construct)


def _represent_ordered(dumper, data):
    return yaml.nodes.MappingNode(
        data.yaml_tag,
        [
            (
                dumper.represent_data(key),
                dumper.represent_data(getattr(data, key))
            )
            for key in data.attr_order
        ]
    )

def _add_data_type(tag, class_):
    class_.yaml_tag = f'!{tag}'
    try:
        constructed(class_)
    except AttributeError:
        class_.yaml_loader = yaml.SafeLoader
        new_class = type(f'Yaml{class_.__name__}', (yaml.YAMLObject,), {k: v for k, v in class_.__dict__.items() if not k.startswith('__')})
        if hasattr(new_class, 'attr_order'):
            ordered(new_class)



import re

from .cond_extractor import cast_conds

STRING_REGEX = r'^(\'[^\']*\')|("[^"]*")$'


def try_cast(data): # pylint: disable=too-many-return-statements
    if data is None:
        return data
    if isinstance(data, list):
        return [try_cast(val) for val in data]
    if isinstance(data, dict):
        return ensure_data_types(data)
    if re.match(STRING_REGEX, data):
        return data[1:-1]
    if data.lower() in ['true', 'yes']:
        return True
    if data.lower() in ['false', 'no']:
        return False
    if data.lower() in ['null', 'nil', 'none']:
        return None

    try:
        return float(data)
    except ValueError as _:
        return data


def ensure_data_types(data):
    return cast_conds({ key: try_cast(val) for key, val in data.items() })
